@@ -167,6 +167,7 @@ import Hledger.Utils
 import Prelude hiding (getContents, writeFile)
 import Hledger.Data.JournalChecks (journalStrictChecks)
 import Text.Printf (printf)
+import Hledger.Data.Journal (journalNumberTransactions)
 
 --- ** doctest setup
 -- $setup
@@ -223,7 +224,7 @@ defaultJournalPath = do
     exists <- doesFileExist defaultfile
     if exists then return defaultfile
     -- else error' $ "LEDGER_FILE is unset and \"" <> defaultfile <> "\" was not found"
-    else error' $ "neither LEDGER_FILE nor \"" <> defaultfile <> "\" were found"
+    else error' $ "neither LEDGER_FILE nor \"" <> defaultfile <> "\" was found"
   else do
     mf <- headMay <$> expandGlob "." ledgerfile `C.catch` (\(_::C.IOException) -> return [])
     case mf of
@@ -350,10 +351,12 @@ readJournalFiles iopts@InputOpts{strict_, new_, new_save_} prefixedfiles = do
 -- The implementation of readJournalFiles.
 -- With --new, it also returns the latest transaction date(s) read in each file
 -- (used by the import command).
+-- This also renumbers the transactions, ensuring their tindex values are unique;
+-- that's also done elsewhere, but some code (accountTransactionsReport) needs it done sooner.
 readJournalFilesAndLatestDates :: InputOpts -> [PrefixedFilePath] -> ExceptT String IO (Journal, [LatestDatesForFile])
 readJournalFilesAndLatestDates iopts pfs = do
   (js, lastdates) <- unzip <$> mapM (readJournalFileAndLatestDates iopts) pfs
-  return (maybe def sconcat $ nonEmpty js, catMaybes lastdates)
+  return (journalNumberTransactions $ maybe def sconcat $ nonEmpty js, catMaybes lastdates)
 
 -- | An easy version of 'readJournal' which assumes default options, and fails in the IO monad.
 readJournal' :: Handle -> IO Journal
